@@ -1,13 +1,13 @@
 ###############################################################################
 #
-# This file copyright (c) 2001-2008 Randy J. Ray, all rights reserved
+# This file copyright (c) 2001-2010 Randy J. Ray, all rights reserved
 #
-# See "LICENSE" in the documentation for licensing and redistribution terms.
+# Copying and distribution are permitted under the terms of the Artistic
+# License 2.0 (http://www.opensource.org/licenses/artistic-license-2.0.php) or
+# the GNU LGPL (http://www.opensource.org/licenses/lgpl-2.1.php).
 #
 ###############################################################################
 #
-#   $Id: Procedure.pm 343 2008-04-09 09:54:36Z rjray $
-#
 #   Description:    This class abstracts out all the procedure-related
 #                   operations from the RPC::XML::Server class
 #
@@ -38,17 +38,26 @@
 
 package RPC::XML::Procedure;
 
-use 5.005;
+use 5.006001;
 use strict;
+use warnings;
 use vars qw($VERSION);
 use subs qw(new is_valid name code signature help version hidden
-            add_signature delete_signature make_sig_table match_signature
-            reload load_XPL_file);
+    add_signature delete_signature make_sig_table match_signature
+    reload load_XPL_file);
 
 use AutoLoader 'AUTOLOAD';
-require File::Spec;
+use File::Spec;
 
-$VERSION = '1.15';
+use Scalar::Util 'blessed';
+
+use RPC::XML 'smart_encode';
+
+# This module also provides RPC::XML::Method
+## no critic (ProhibitMultiplePackages)
+
+$VERSION = '1.21';
+$VERSION = eval $VERSION;    ## no critic (ProhibitStringyEval)
 
 ###############################################################################
 #
@@ -68,10 +77,9 @@ $VERSION = '1.15';
 ###############################################################################
 sub new
 {
-    my $class = shift;
-    my @argz  = @_;
+    my ($class , @argz) = @_;
 
-    my $data; # This will be a hashref that eventually gets blessed
+    my $data;    # This will be a hashref that eventually gets blessed
 
     $class = ref($class) || $class;
 
@@ -81,8 +89,8 @@ sub new
     if (ref $argz[0])
     {
         # 1. A hashref containing all the relevant keys
-        $data = {};
-        %$data = %{$argz[0]};
+        $data  = {};
+        %{$data} = %{$argz[0]};
     }
     elsif (@argz == 1)
     {
@@ -99,14 +107,22 @@ sub new
         if (defined $class)
         {
             $data = $class->load_XPL_file($argz[0]);
-            return $data unless ref $data; # load_XPL_path signalled an error
+            if (! ref $data)
+            {
+                # load_XPL_path signalled an error
+                return $data;
+            }
         }
         else
         {
             # Spoofing the "class" argument to load_XPL_file makes me feel
             # even dirtier...
             $data = load_XPL_file(\$class, $argz[0]);
-            return $data unless ref $data; # load_XPL_path signalled an error
+            if (! ref $data)
+            {
+                # load_XPL_path signalled an error
+                return $data;
+            }
             $class = "RPC::XML::$class";
         }
     }
@@ -119,14 +135,14 @@ sub new
         $data->{signature} = [];
         while (@argz)
         {
-            ($key, $val) = splice(@argz, 0, 2);
+            ($key, $val) = splice @argz, 0, 2;
             if ($key eq 'signature')
             {
                 # Since there may be more than one signature, we allow it to
                 # repeat. Of course, that's also why we can't just take @argz
                 # directly as a hash. *shrug*
-                push(@{$data->{signature}},
-                     ref($val) ? join(' ', @$val) : $val);
+                push @{$data->{signature}},
+                    ref $val ? join q{ } => @{$val} : $val;
             }
             elsif (exists $data->{$key})
             {
@@ -139,14 +155,18 @@ sub new
         }
     }
 
-    return "${class}::new: Missing required data"
-        unless (exists $data->{signature} and
-                (ref($data->{signature}) eq 'ARRAY') and
-                scalar(@{$data->{signature}}) and
-                $data->{name} and $data->{code});
+    if (! (exists $data->{signature} &&
+           (ref($data->{signature}) eq 'ARRAY') &&
+           scalar(@{$data->{signature}}) &&
+           $data->{name} &&
+           $data->{code}))
+    {
+        return "${class}::new: Missing required data";
+    }
     bless $data, $class;
+
     # This needs to happen post-bless in case of error (for error messages)
-    $data->make_sig_table;
+    return $data->make_sig_table;
 }
 
 ###############################################################################
@@ -168,63 +188,107 @@ sub make_sig_table
 {
     my $self = shift;
 
-    my ($sig, $return, $rest);
+    my ($return, $rest);
 
     delete $self->{sig_table};
-    for $sig (@{$self->{signature}})
+    for my $sig (@{$self->{signature}})
     {
-        ($return, $rest) = split(/ /, $sig, 2); $rest = '' unless $rest;
+        ($return, $rest) = split / /, $sig, 2;
+        if (! $rest)
+        {
+            $rest = q{};
+        }
         # If the key $rest already exists, then this is a collision
-        return ref($self) . '::make_sig_table: Cannot have two different ' .
-            "return values for one set of params ($return vs. " .
-            "$self->{sig_table}->{$rest})"
-                if $self->{sig_table}->{$rest};
+        if ($self->{sig_table}->{$rest})
+        {
+            return
+                ref($self) . '::make_sig_table: Cannot have two different ' .
+                "return values for one set of params ($return vs. " .
+                "$self->{sig_table}->{$rest})";
+        }
+
         $self->{sig_table}->{$rest} = $return;
     }
 
-    $self;
+    return $self;
 }
 
 #
 # These are basic accessor/setting functions for the various attributes
 #
-sub name      { $_[0]->{name}; } # "name" cannot be changed at this level
-sub help      { $_[1] and $_[0]->{help}    = $_[1]; $_[0]->{help};    }
-sub version   { $_[1] and $_[0]->{version} = $_[1]; $_[0]->{version}; }
-sub hidden    { $_[1] and $_[0]->{hidden}  = $_[1]; $_[0]->{hidden};  }
+sub name { return shift->{name}; }    # "name" cannot be changed at this level
+sub namespace { return shift->{namespace} || q{}; }    # Nor can "namespace"
+
+sub help
+{
+    my ($self, $value) = @_;
+
+    $value and $self->{help} = $value;
+
+    return $self->{help};
+}
+
+sub version
+{
+    my ($self, $value) = @_;
+
+    $value and $self->{version} = $value;
+
+    return $self->{version};
+}
+
+sub hidden
+{
+    my ($self, $value) = @_;
+
+    $value and $self->{hidden} = $value;
+
+    return $self->{hidden};
+}
+
 sub code
 {
-    ref $_[1] eq 'CODE' and $_[0]->{code} = $_[1];
-    $_[0]->{code};
+    my ($self, $value) = @_;
+
+    ref $value eq 'CODE' and $self->{code} = $value;
+
+    return $self->{code};
 }
+
 sub signature
 {
-    if ($_[1] and ref $_[1] eq 'ARRAY')
+    my ($self, $sig) = @_;
+
+    if ($sig and ref $sig eq 'ARRAY')
     {
-        my $old = $_[0]->{signature};
-        $_[0]->{signature} = $_[1];
-        unless (ref($_[0]->make_sig_table))
+        my $old = $self->{signature};
+        $self->{signature} = $sig;
+        if (! ref $self->make_sig_table)
         {
             # If it failed to re-init the table, restore the old list (and old
             # table). We don't have to check this return, since it had worked
-            $_[0]->{signature} = $old;
-            $_[0]->make_sig_table;
+            # before.
+            $self->{signature} = $old;
+            $self->make_sig_table;
         }
     }
+
     # Return a copy of the array, not the original
-    [ @{$_[0]->{signature}} ];
+    return [ @{$self->{signature}} ];
 }
 
 package RPC::XML::Method;
 
 use strict;
+use base qw(RPC::XML::Procedure);
 
-@RPC::XML::Method::ISA = qw(RPC::XML::Procedure);
-
+# Needed for AutoSplit to be happy:
 package RPC::XML::Procedure;
 
 1;
 
+__END__
+
 =head1 NAME
 
 RPC::XML::Procedure - Object encapsulation of server-side RPC procedures
@@ -265,7 +329,7 @@ clear, intuitive hash keys (rather than obfuscated as the other classes
 do). Thus it is important to be clear on the interface here before
 sub-classing this package.
 
-=head1 USAGE
+=head1 SUBROUTINES/METHODS
 
 The following methods are provided by this class:
 
@@ -356,6 +420,14 @@ persist. However, the other elements may be used in the creation of a new
 object, which may then be added to the server, if the name absolutely must
 change.
 
+=item namespace
+
+If the procedure object was created from a file, or if the instantiation
+included namespace information, this accessor will return the namespace that
+the underlying code executes in. Otherwise, it returns an empty string. This
+cannot be altered (even if the B<code> method is used to replace the code
+routine).
+
 =item code([NEW])
 
 Returns or sets the code-reference that will receive calls as marshalled by
@@ -457,6 +529,14 @@ protected, and the same care should be taken before altering any of them:
 When the method was loaded from a file, this key contains the path to the file
 used.
 
+=item namespace
+
+If the code is loaded from a file, this hash key will reflect what namespace
+the code executes in. If the file specified a namespace, that is the value
+you will get (any occurrence of C<.> in the specified namespace will have been
+converted to C<::>). If no explicit namespace was provided, the namespace
+of the class you called B<new> from will be used. See L</"Namespaces">.
+
 =item mtime
 
 When the method was loaded from a file, this key contains the
@@ -493,11 +573,12 @@ other packages than this one, or useful in other contexts than this one.
 
 The lightweight DTD for the layout can be summarized as:
 
-        <!ELEMENT  proceduredef  (name, version?, hidden?, signature+,
-                                  help?, code)>
-        <!ELEMENT  methoddef  (name, version?, hidden?, signature+,
-                               help?, code)>
+        <!ELEMENT  proceduredef  (name, namespace?, version?, hidden?,
+                                  signature+, help?, code)>
+        <!ELEMENT  methoddef     (name, namespace?, version?, hidden?,
+                                  signature+, help?, code)>
         <!ELEMENT  name       (#PCDATA)>
+        <!ELEMENT  namespace  (#PCDATA)>
         <!ELEMENT  version    (#PCDATA)>
         <!ELEMENT  hidden     EMPTY>
         <!ELEMENT  signature  (#PCDATA)>
@@ -505,8 +586,8 @@ The lightweight DTD for the layout can be summarized as:
         <!ELEMENT  code       (#PCDATA)>
         <!ATTLIST  code       language (#PCDATA)>
 
-The containing tag is always one of C<E<lt>methoddefE<gt>> or
-C<E<lt>proceduredefE<gt>>. The tags that specify name, signatures and the code
+The containing tag is always one of C<< <methoddef> >> or
+C<< <proceduredef> >>. The tags that specify name, signatures and the code
 itself must always be present. Some optional information may also be
 supplied. The "help" text, or what an introspection API would expect to use to
 document the method, is also marked as optional.  Having some degree of
@@ -515,7 +596,7 @@ however.
 
 The default methods that this package provides are turned into XPL files by
 the B<make_method> tool (see L<make_method>). The final forms of these may
-serve as direct examples of what the file should look like.
+serve as examples of what the file should look like.
 
 =item Information used only for book-keeping
 
@@ -531,16 +612,16 @@ to do so.
 =item The information crucial to the method
 
 The name, signatures and code must be present for obvious reasons. The
-C<E<lt>nameE<gt>> tag tells the server what external name this procedure is
-known by. The C<E<lt>signatureE<gt>> tag, which may appear more than once,
+C<< <name> >> tag tells the server what external name this procedure is
+known by. The C<< <signature> >> tag, which may appear more than once,
 provides the definition of the interface to the function in terms of what
 types and quantity of arguments it will accept, and for a given set of
 arguments what the type of the returned value is. Lastly is the
-C<E<lt>codeE<gt>> tag, without which there is no procedure to remotely call.
+C<< <code> >> tag, without which there is no procedure to remotely call.
 
 =item Why the <code> tag allows multiple languages
 
-Note that the C<E<lt>codeE<gt>> tag is the only one with an attribute, in this
+Note that the C<< <code> >> tag is the only one with an attribute, in this
 case "language". This is designed to allow for one XPL file to provide a given
 method in multiple languages. Why, one might ask, would there be a need for
 this?
@@ -556,7 +637,9 @@ implementations in all four of the above languages, the name, help text,
 signature and even hidden status would likely be identical. So, why not share
 the non-language-specific elements in the spirit of re-use?
 
-=item The "make_method" utility
+=back
+
+=head2 The C<make_method> Utility
 
 The utility script C<make_method> is provided as a part of this software
 suite. It allows for the automatic creation of XPL files from either
@@ -567,7 +650,26 @@ automatic generation of XPL files and their delivery as a part of the normal
 Perl module build process. Using this tool is highly recommended over managing
 XPL files directly. For the full details, see L<make_method>.
 
-=back
+=head1 NAMESPACES
+
+As default behavior, Perl code that is passed to C<eval> when a XPL file is
+loaded gets put into the same namespace as the package used to load the XPL.
+It is not an issue when you create your own B<RPC::XML::Procedure> (or
+B<::Method> or B<::Function>) objects, as the code is already instantiated
+into a given namespace.  This can be important if your code expects to call
+routines in other loaded packages, utilize package-level globals, etc.
+
+To give developers control over the namespace in XPL code, a new optional
+tag C<< <namespace> >> was added in the 0.65 release. If this tag is present
+in the XPL being read, it defines the namespace that the C<< <code> >> block
+is evaluated in.
+
+The value of the namespace tag is a string providing the namespace in either
+the Perl-style of hierarchy parts separated by C<::>, or the style used by
+Java, Perl6, etc., in which the parts are separated by C<.>. The latter
+form is converted to Perl style for the evaluation of the code. If there is
+no namespace declaration in a XPL file, the namespace of the class that
+loads the XPL is used.
 
 =head1 DIAGNOSTICS
 
@@ -581,14 +683,53 @@ Moving the method management to a separate class adds a good deal of overhead
 to the general system. The trade-off in reduced complexity and added
 maintainability should offset this.
 
-=head1 LICENSE
+=head1 BUGS
 
-This module and the code within are released under the terms of the Artistic
-License 2.0
-(http://www.opensource.org/licenses/artistic-license-2.0.php). This code may
-be redistributed under either the Artistic License or the GNU Lesser General
-Public License (LGPL) version 2.1
-(http://www.opensource.org/licenses/lgpl-license.php).
+Please report any bugs or feature requests to
+C<bug-rpc-xml at rt.cpan.org>, or through the web interface at
+L<http://rt.cpan.org/NoAuth/ReportBug.html?Queue=RPC-XML>. I will be
+notified, and then you'll automatically be notified of progress on
+your bug as I make changes.
+
+=head1 SUPPORT
+
+=over 4
+
+=item * RT: CPAN's request tracker
+
+L<http://rt.cpan.org/NoAuth/Bugs.html?Dist=RPC-XML>
+
+=item * AnnoCPAN: Annotated CPAN documentation
+
+L<http://annocpan.org/dist/RPC-XML>
+
+=item * CPAN Ratings
+
+L<http://cpanratings.perl.org/d/RPC-XML>
+
+=item * Search CPAN
+
+L<http://search.cpan.org/dist/RPC-XML>
+
+=item * Source code on GitHub
+
+L<http://github.com/rjray/rpc-xml>
+
+=back
+
+=head1 LICENSE AND COPYRIGHT
+
+This file and the code within are copyright (c) 2010 by Randy J. Ray.
+
+Copying and distribution are permitted under the terms of the Artistic
+License 2.0 (L<http://www.opensource.org/licenses/artistic-license-2.0.php>) or
+the GNU LGPL 2.1 (L<http://www.opensource.org/licenses/lgpl-2.1.php>).
+
+=head1 CREDITS
+
+The B<XML-RPC> standard is Copyright (c) 1998-2001, UserLand Software, Inc.
+See <http://www.xmlrpc.com> for more information about the B<XML-RPC>
+specification.
 
 =head1 SEE ALSO
 
@@ -596,12 +737,10 @@ L<RPC::XML::Server>, L<make_method>
 
 =head1 AUTHOR
 
-Randy J. Ray <rjray@blackperl.com>
+Randy J. Ray C<< <rjray@blackperl.com> >>
 
 =cut
 
-__END__
-
 ###############################################################################
 #
 #   Sub Name:       clone
@@ -622,7 +761,7 @@ sub clone
     my $self = shift;
 
     my $new_self = {};
-    for (keys %$self)
+    for (keys %{$self})
     {
         next if $_ eq 'signature';
         $new_self->{$_} = $self->{$_};
@@ -630,7 +769,7 @@ sub clone
     $new_self->{signature} = [];
     @{$new_self->{signature}} = @{$self->{signature}};
 
-    bless $new_self, ref($self);
+    return bless $new_self, ref $self;
 }
 
 ###############################################################################
@@ -652,8 +791,9 @@ sub is_valid
 {
     my $self = shift;
 
-    return ((ref($self->{code}) eq 'CODE') and $self->{name} and
-            (ref($self->{signature}) && scalar(@{$self->{signature}})));
+    return (    (ref($self->{code}) eq 'CODE')
+            and $self->{name}
+            and (ref($self->{signature}) && scalar(@{$self->{signature}})));
 }
 
 ###############################################################################
@@ -674,60 +814,59 @@ sub is_valid
 ###############################################################################
 sub add_signature
 {
-    my $self = shift;
-    my @args = @_;
+    my ($self, @args) = @_;
 
-    my (%sigs, $one_sig, $tmp, $old);
+    my (%sigs, $tmp, $old);
 
     # Preserve the original in case adding the new one causes a problem
     $old = $self->{signature};
     %sigs = map { $_ => 1 } @{$self->{signature}};
-    for $one_sig (@args)
+    for my $one_sig (@args)
     {
-        $tmp = (ref $one_sig) ? join(' ', @$one_sig) : $one_sig;
+        $tmp = (ref $one_sig) ? join q{ } => @{$one_sig} : $one_sig;
         $sigs{$tmp} = 1;
     }
     $self->{signature} = [ keys %sigs ];
-    unless (ref($tmp = $self->make_sig_table))
+    $tmp = $self->make_sig_table;
+    if (! ref $tmp)
     {
         # Because this failed, we have to restore the old table and return
         # an error
         $self->{signature} = $old;
         $self->make_sig_table;
-        return ref($self) . '::add_signature: Error re-hashing table: ' .
-            $tmp;
+        return ref($self) . '::add_signature: Error re-hashing table: ' . $tmp;
     }
 
-    $self;
+    return $self;
 }
 
 sub delete_signature
 {
-    my $self = shift;
-    my @args = @_;
+    my ($self, @args) = @_;
 
-    my (%sigs, $one_sig, $tmp, $old);
+    my (%sigs, $tmp, $old);
 
     # Preserve the original in case adding the new one causes a problem
     $old = $self->{signature};
     %sigs = map { $_ => 1 } @{$self->{signature}};
-    for $one_sig (@args)
+    for my $one_sig (@args)
     {
-        $tmp = (ref $one_sig) ? join(' ', @$one_sig) : $one_sig;
+        $tmp = (ref $one_sig) ? join q{ } => @{$one_sig} : $one_sig;
         delete $sigs{$tmp};
     }
     $self->{signature} = [ keys %sigs ];
-    unless (ref($tmp = $self->make_sig_table))
+    $tmp = $self->make_sig_table;
+    if  (! ref $tmp)
     {
         # Because this failed, we have to restore the old table and return
         # an error
         $self->{signature} = $old;
         $self->make_sig_table;
-        return ref($self) . '::delete_signature: Error re-hashing table: ' .
-            $tmp;
+        return
+            ref $self . '::delete_signature: Error re-hashing table: ' . $tmp;
     }
 
-    $self;
+    return $self;
 }
 
 ###############################################################################
@@ -750,7 +889,10 @@ sub match_signature
     my $self = shift;
     my $sig  = shift;
 
-    $sig = join(' ', @$sig) if ref $sig;
+    if (ref $sig)
+    {
+        $sig = join q{ } => @{$sig};
+    }
 
     return $self->{sig_table}->{$sig} || 0;
 }
@@ -772,14 +914,21 @@ sub reload
 {
     my $self = shift;
 
-    return ref($self) . '::reload: No file associated with method ' .
-        $self->{name} unless $self->{file};
+    if (! $self->{file})
+    {
+        return sprintf '%s::reload: No file associated with method %s',
+            ref $self, $self->{name};
+    }
+
     my $tmp = $self->load_XPL_file($self->{file});
 
     if (ref $tmp)
     {
         # Update the information on this actual object
-        $self->{$_} = $tmp->{$_} for (keys %$tmp);
+        for (keys %{$tmp})
+        {
+            $self->{$_} = $tmp->{$_};
+        }
         # Re-calculate the signature table, in case that changed as well
         return $self->make_sig_table;
     }
@@ -814,8 +963,7 @@ sub load_XPL_file
 
     require XML::Parser;
 
-    my ($me, $pkg, $data, $signature, $code, $codetext, $accum, $P, %attr);
-    local *F;
+    my ($me, $data, $signature, $code, $codetext, $accum, $P, %attr);
 
     if (ref($self) eq 'SCALAR')
     {
@@ -828,64 +976,96 @@ sub load_XPL_file
     }
     $data = {};
     # So these don't end up undef, since they're optional elements
-    $data->{hidden} = 0; $data->{version} = ''; $data->{help} = '';
-    $data->{called} = 0;
-    open(F, "< $file") or return "$me: Error opening $file for reading: $!";
-    $P = XML::Parser
-        ->new(Handlers => {Char  => sub { $accum .= $_[1] },
-                           Start => sub { %attr = splice(@_, 2) },
-                           End   =>
-                           sub {
-                               my $elem = $_[1];
+    $data->{hidden}  = 0;
+    $data->{version} = q{};
+    $data->{help}    = q{};
+    $data->{called}  = 0;
+    $P = XML::Parser->new(
+        ErrorContext => 1,
+        Handlers     => {
+            Char => sub { $accum .= $_[1] },
+            Start => sub { %attr = splice @_, 2 },
+            End => sub {
+                my $elem = $_[1];
 
-                               $accum =~ s/^[\s\n]+//;
-                               $accum =~ s/[\s\n]+$//;
-                               if ($elem eq 'signature')
-                               {
-                                   $data->{signature} ||= [];
-                                   push(@{$data->{signature}}, $accum);
-                               }
-                               elsif ($elem eq 'code')
-                               {
-                                   $data->{$elem} = $accum
-                                       unless ($attr{language} and
-                                               $attr{language} ne 'perl');
-                               }
-                               elsif (substr($elem, -3) eq 'def')
-                               {
-                                   # Don't blindly store the container tag...
-                                   # We may need it to tell the caller what
-                                   # our type is
-                                   $$self = ucfirst substr($elem, 0, -3)
-                                       if (ref($self) eq 'SCALAR');
-                               }
-                               else
-                               {
-                                   $data->{$elem} = $accum;
-                               }
+                $accum =~ s/^[\s\n]+//;
+                $accum =~ s/[\s\n]+$//;
+                if ($elem eq 'signature')
+                {
+                    $data->{signature} ||= [];
+                    push @{$data->{signature}}, $accum;
+                }
+                elsif ($elem eq 'code')
+                {
+                    if (! ($attr{language} &&
+                           $attr{language} ne 'perl'))
+                    {
+                        $data->{$elem} = $accum;
+                    }
+                }
+                elsif ('def' eq substr $elem, -3)
+                {
+                    # Don't blindly store the container tag...
+                    # We may need it to tell the caller what
+                    # our type is
+                    if (ref $self eq 'SCALAR')
+                    {
+                        ${$self} = ucfirst substr $elem, 0, -3;
+                    }
+                }
+                else
+                {
+                    $data->{$elem} = $accum;
+                }
 
-                               %attr = ();
-                               $accum = '';
-                           }});
-    return "$me: Error creating XML::Parser object" unless $P;
+                %attr  = ();
+                $accum = q{};
+            }
+        }
+    );
+    if (! $P)
+    {
+        return "$me: Error creating XML::Parser object";
+    }
+    open my $fh, '<', $file or
+        return "$me: Error opening $file for reading: $!";
     # Trap any errors
-    eval { $P->parse(*F) };
-    close(F);
-    return "$me: Error parsing $file: $@" if $@;
+    eval { $P->parse($fh); }; ## no critic (RequireCheckingReturnValueOfEval)
+    close $fh or return "$me: Error closing $file: $!";
+    if ($@)
+    {
+        return "$me: Error parsing $file: $@";
+    }
 
     # Try to normalize $codetext before passing it to eval
-    my $class = __PACKAGE__; # token won't expand in the s/// below
+
+    # First step is set the namespace the code will live in. The default is
+    # the package that we're in (be it ::Procedure, ::Method, etc.). If they
+    # specify one, use it instead.
+    if ($data->{namespace})
+    {
+        # Fudge a little and let them '.' as a synonym for '::' in the
+        # namespace hierarchy.
+        $data->{namespace} =~ s{\.}{::}g;
+    }
+    else
+    {
+        $data->{namespace} = __PACKAGE__;
+    }
+
+    # Next step is to munge away any actual subroutine name so that the eval
+    # yields an anonymous sub. Also insert the namespace declaration.
     ($codetext = $data->{code}) =~
-        s/sub[\s\n]+([\w:]+)?[\s\n]*\{/sub \{ package $class; /;
-    $code = eval $codetext;
+        s/sub[\s\n]+([\w:]+)?[\s\n]*\{/sub \{ package $data->{namespace}; /x;
+    $code = eval $codetext; ## no critic (ProhibitStringyEval)
     return "$me: Error creating anonymous sub: $@" if $@;
 
     $data->{code} = $code;
     # Add the file's mtime for when we check for stat-based reloading
     $data->{mtime} = (stat $file)[9];
-    $data->{file} = $file;
+    $data->{file}  = $file;
 
-    $data;
+    return $data;
 }
 
 ###############################################################################
@@ -919,46 +1099,80 @@ sub call
     $name = $self->name;
     # Create the param list.
     # The type for the response will be derived from the matching signature
-    @paramtypes = map { $_->type  } @data;
+    @paramtypes = map { $_->type } @data;
     @params     = map { $_->value } @data;
-    $signature = join(' ', @paramtypes);
+    $signature = join q{ } => @paramtypes;
     $resptype = $self->match_signature($signature);
     # Since there must be at least one signature with a return value (even
     # if the param list is empty), this tells us if the signature matches:
-    return RPC::XML::fault->new(301,
-                                "method $name has no matching " .
-                                'signature for the argument list: ' .
-                                "[$signature]")
-        unless ($resptype);
+    if (! $resptype)
+    {
+        return $srv->server_fault(
+            badsignature =>
+            "method $name has no matching signature for the argument list: " .
+            "[$signature]"
+        );
+    }
+
+    # Make sure that the response-type is a valid XML-RPC type
+    if (($resptype ne 'scalar') && (! "RPC::XML::$resptype"->can('new')))
+    {
+        return $srv->server_fault(badsignature =>
+            "Signature [$signature] for method $name has unknown " .
+            "return-type '$resptype'");
+    }
 
     # Set these in case the server object is part of the param list
-    local $srv->{signature} = [ $resptype, @paramtypes ];
-    local $srv->{method_name} = $name;
+    local $srv->{signature} =          ## no critic (ProhibitLocalVars)
+        [ $resptype, @paramtypes ];
+    local $srv->{method_name} = $name; ## no critic (ProhibitLocalVars)
     # If the method being called is "system.status", check to see if we should
     # increment the server call-count.
-    $noinc = (($name eq 'system.status') && @data &&
-              ($paramtypes[0] eq 'boolean') && $params[0]) ? 1 : 0;
+    $noinc =
+        (($name eq 'system.status') &&
+            @data &&
+            ($paramtypes[0] eq 'boolean') &&
+            $params[0]) ? 1 : 0;
     # For RPC::XML::Method (and derivatives), pass the server object
-    unshift(@params, $srv) if ($self->isa('RPC::XML::Method'));
+    if ($self->isa('RPC::XML::Method'))
+    {
+        unshift @params, $srv;
+    }
 
     # Now take a deep breath and call the method with the arguments
-    eval { $response = $self->{code}->(@params); };
-    # On failure, propagate user-generated RPC::XML::fault exceptions, or
-    # transform Perl-level error/failure into such an object
-    if ($@)
+    if (! eval { $response = $self->{code}->(@params); 1; })
     {
-        return UNIVERSAL::isa($@, 'RPC::XML::fault') ?
-            $@ :
-            RPC::XML::fault->new(302, "Method $name returned error: $@");
+        # On failure, propagate user-generated RPC::XML::fault exceptions, or
+        # transform Perl-level error/failure into such an object
+        if ($@)
+        {
+            return (blessed $@ and $@->isa('RPC::XML::fault')) ?
+                $@ : $srv->server_fault(execerror =>
+                                        "Method $name returned error: $@");
+        }
     }
 
-    $self->{called}++ unless $noinc;
+    if (! $noinc)
+    {
+        $self->{called}++;
+    }
     # Create a suitable return value
-    if ((! ref($response)) && UNIVERSAL::can("RPC::XML::$resptype", 'new'))
+    if (! ref $response)
     {
-        my $class = "RPC::XML::$resptype";
-        $response = $class->new($response);
+        if ($resptype eq 'scalar')
+        {
+            # Server code from the RPC::XML::Function class doesn't use
+            # signatures, so if they didn't encode the returned value
+            # themselves they're trusting smart_encode() to get it right.
+            $response = smart_encode($response);
+        }
+        else
+        {
+            # We checked that this was valid earlier, so no need for further
+            # tests here.
+            $response = "RPC::XML::$resptype"->new($response);
+        }
     }
 
-    $response;
+    return $response;
 }
